#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re
import subprocess
import unittest

from distutils.version import LooseVersion

from . import PelicanPluginTestCase

def dot_version():
    return re.match(".*version (?P<version>\d+\.\d+\.\d+).*", subprocess.check_output(['dot', '-V'], stderr=subprocess.STDOUT).decode('utf-8').strip()).group('version')

class Dot(PelicanPluginTestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(__file__, '', *args, **kwargs)

    def test(self):
        self.run_pelican({
            'PLUGINS': ['m.htmlsanity', 'm.components', 'm.dot'],
            'M_DOT_FONT': 'DejaVu Sans'
        })

        # Used to be >= 2.44.0, but 2.42.2 appears to have the same output
        if LooseVersion(dot_version()) >= LooseVersion("2.42.2"):
            file = 'page.html'
        else:
            file = 'page-240.html'

        self.assertEqual(*self.actual_expected_contents('page.html', file))



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

from pygments.style import Style
from pygments.token import Text, Generic

class ConsoleStyle(Style):

    background_color = None
    default_style = ''

    styles = {
        Text:                                 '#fcfcfc',

        # Shell session
        Generic.Prompt:                       'bold #16a085',  # same as BrightCyan
        Generic.Output:                       '#fcfcfc',

        # ANSI highlighting. Same order as in Konsole style dialog, following
        # the Breeze theme.
        Generic.AnsiBlack:                    '#232627',
        Generic.AnsiRed:                      '#ed1515',
        Generic.AnsiGreen:                    '#11d116',
        Generic.AnsiYellow:                   '#f67400',
        Generic.AnsiBlue:                     '#1d99f3',
        Generic.AnsiMagenta:                  '#9b59b6',
        Generic.AnsiCyan:                     '#1abc9c',
        Generic.AnsiWhite:                    '#fcfcfc',
        Generic.AnsiBrightBlack:              'bold #7f8c8d',
        Generic.AnsiBrightRed:                'bold #c0392b',
        Generic.AnsiBrightGreen:              'bold #1cdc9a',
        Generic.AnsiBrightYellow:             'bold #fdbc4b',
        Generic.AnsiBrightBlue:               'bold #3daee9',
        Generic.AnsiBrightMagenta:            'bold #8e44ad',
        Generic.AnsiBrightCyan:               'bold #16a085',
        Generic.AnsiBrightWhite:              'bold #ffffff',
        Generic.AnsiBrightDefault:            'bold #ffffff',
        Generic.AnsiBackgroundBlack:          'bg:#232627',
        Generic.AnsiBackgroundRed:            'bg:#ed1515',
        Generic.AnsiBackgroundGreen:          'bg:#11d116',
        Generic.AnsiBackgroundYellow:         'bg:#f67400',
        Generic.AnsiBackgroundBlue:           'bg:#1d99f3',
        Generic.AnsiBackgroundMagenta:        'bg:#9b59b6',
        Generic.AnsiBackgroundCyan:           'bg:#1abc9c',
        Generic.AnsiBackgroundWhite:          'bg:#fcfcfc',
        Generic.AnsiBackgroundBrightBlack:    'bg:#7f8c8d',
        Generic.AnsiBackgroundBrightRed:      'bg:#c0392b',
        Generic.AnsiBackgroundBrightGreen:    'bg:#1cdc9a',
        Generic.AnsiBackgroundBrightYellow:   'bg:#fdbc4b',
        Generic.AnsiBackgroundBrightBlue:     'bg:#3daee9',
        Generic.AnsiBackgroundBrightMagenta:  'bg:#8e44ad',
        Generic.AnsiBackgroundBrightCyan:     'bg:#16a085',
        Generic.AnsiBackgroundBrightWhite:    'bg:#ffffff',
    }



#!/usr/bin/env python3

#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import argparse
import re
import os
import sys

import_rx = re.compile("^@import url\\('(?P<file>[^']+)'\\);$")
opening_brace_rx = re.compile("^\\s*:root\s*{\\s*$")
closing_brace_rx = re.compile("^\\s*}\\s*$")
comment_rx = re.compile("^\\s*(/\\*.*\\*/)?\\s*$")
comment_start_rx = re.compile("^\\s*(/\\*.*)\\s*$")
comment_end_rx = re.compile("^\\s*(.*\\*/)\\s*$")
variable_declaration_rx = re.compile("^\\s*(?P<key>--[a-z-]+)\\s*:\\s*(?P<value>[^;]+)\\s*;\\s*(/\\*.*\\*/)?\\s*$")
variable_use_rx = re.compile("^(?P<before>.*)var\\((?P<key>--[a-z-]+)\\)(?P<after>.*)$")

def postprocess(files, process_imports, out_file):
    directory = os.path.dirname(files[0])

    if not out_file:
        basename, ext = os.path.splitext(files[0])
        out_file = basename + ".compiled" + ext

    variables = {}
    imported_files = []
    def parse(f):
        nonlocal variables, imported_files
        not_just_variable_declarations = False
        in_variable_declarations = False
        in_comment = False
        for line in f:
            # In comment and the comment is not ending yet, ignore
            if in_comment:
                if comment_end_rx.match(line):
                    in_comment = False
                continue

            # Import statement: add the file to additionally processed files
            # unless it's disabled
            match = import_rx.match(line)
            if match:
                if process_imports:
                    imported_files += [match.group('file')]
                continue

            # Opening brace of variable declaration block
            match = opening_brace_rx.match(line)
            if match:
                in_variable_declarations = True
                continue

            # Variable declaration
            match = variable_declaration_rx.match(line)
            if match and in_variable_declarations:
                # Define a variable to equal another
                key = match.group('key')
                value = match.group('value')
                match = variable_use_rx.match(value)
                if match and match.group('key') in variables:
                    value = match.group('before') + variables[match.group('key')]
                    if match.group('after').endswith('*/'):
                        value += match.group('after')[:match.group('after').rindex('/*')].rstrip()
                    else:
                        value += match.group('after')
                variables[key] = value
                continue

            # Comment or empty line, ignore
            if comment_rx.match(line):
                continue

            # Comment start line, ignore this and the next lines
            if comment_start_rx.match(line):
                in_comment = True
                continue

            # Closing brace of variable declaration block. If it was not just
            # variable declarations, put the closing brace to the output as
            # well.
            match = closing_brace_rx.match(line)
            if match and in_variable_declarations:
                if not_just_variable_declarations: out.write("}\n")
                in_variable_declarations = False
                continue

            # If inside variable declaration block, include also the opening
            # brace and remember to put the closing brace there as well
            if in_variable_declarations:
                out.write(":root {\n")
                not_just_variable_declarations = True

            # Variable use, replace with actual value
            # TODO: more variables on the same line?
            match = variable_use_rx.match(line)
            if match and match.group('key') in variables:
                out.write(match.group('before'))
                out.write(variables[match.group('key')])
                # Strip the trailing comment, if there, to save some bytes
                if match.group('after').endswith('*/'):
                    out.write(match.group('after')[:match.group('after').rindex('/*')].rstrip())
                else:
                    out.write(match.group('after'))
                out.write("\n")
                continue

            # Something else, copy verbatim to the output. Strip the trailing
            # comment, if there, to save some bytes.
            if line.rstrip().endswith('*/'):
                out.write(line[:line.rindex('/*')].rstrip() + '\n')
            else:
                out.write(line)

    with open(out_file, mode='w', encoding='utf8') as out:
        # Put a helper comment and a license blob on top
        out.write("""/* Generated using `./postprocess.py {}`. Do not edit. */

/*
    This file is part of m.css.

    Copyright © 2017, 2018, 2019, 2020, 2021, 2022
              Vladimír Vondruš <mosra@centrum.cz>

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/
""".format(' '.join(sys.argv[1:])))

        # Parse the top-level file
        with open(files[0], encoding='utf8') as f: parse(f)

        # Now open the imported files and parse them as well. Not doing any
        # recursive parsing.
        for i, file in enumerate(imported_files + files[1:]):
            if i: out.write('\n')

            with open(file, encoding='utf8') as f: parse(f)

    return 0

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=r"""
Postprocessor for removing @import statements and variables from CSS files.

Combines all files into a new *.compiled.css file. The basename is taken
implicitly from the first argument. The -o option can override the output
filename.""")
    parser.add_argument('files', nargs='+', help="input CSS file(s)")
    parser.add_argument('--no-import', help="ignore @import statements", action='store_true')
    parser.add_argument('-o', '--output', help="output file", default='')
    args = parser.parse_args()

    exit(postprocess(args.files, not args.no_import, args.output))



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

try:
	from poxy.main import _run
except:
	from main import _run



if __name__ == '__main__':
	_run()



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import os
import shutil
import unittest

from pelican import read_settings, Pelican

class PelicanPluginTestCase(unittest.TestCase):
    def __init__(self, path, dir, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        # Source files for test_something.py are in something_{dir}/ subdirectory
        self.path = os.path.join(os.path.dirname(os.path.realpath(path)), os.path.splitext(os.path.basename(path))[0][5:] + ('_' + dir if dir else ''))

        # Display ALL THE DIFFS
        self.maxDiff = None

    def setUp(self):
        if os.path.exists(os.path.join(self.path, 'output')): shutil.rmtree(os.path.join(self.path, 'output'))

    def run_pelican(self, settings):
        implicit_settings = {
            # Enable this to get proper backtraces
            #'DEBUG': True,

            # Contains just stuff that isn't required by the m.css theme itself,
            # but is needed to have the test setup working correctly
            'RELATIVE_URLS': True,
            'TIMEZONE': 'UTC',
            'READERS': {'html': None},
            'SITEURL': '.',
            'PATH': os.path.join(self.path),
            'OUTPUT_PATH': os.path.join(self.path, 'output'),
            'PAGE_PATHS': [''],
            'PAGE_SAVE_AS': '{slug}.html',
            'PAGE_URL': '{slug}.html',
            'PAGE_EXCLUDES': ['output'],
            'ARTICLE_PATHS': ['articles'], # does not exist
            # Don't render feeds, we don't want to test them all the time
            'FEED_ALL_ATOM': None,
            'CATEGORY_FEED_ATOM': None,
            'THEME': '../pelican-theme',
            'PLUGIN_PATHS': ['.'],
            'THEME_STATIC_DIR': 'static',
            'M_CSS_FILES': ['https://fonts.googleapis.com/css?family=Source+Sans+Pro:400,400i,600,600i',
               'static/m-dark.css'],
            'M_FINE_PRINT': None,
            'M_DISABLE_SOCIAL_META_TAGS': True,
            'DIRECT_TEMPLATES': [],
            'SLUGIFY_SOURCE': 'basename',

            'DOCUTILS_SETTINGS': {
                # Default changed to '%' in 0.18, keep the old setting to
                # have consistent output across versions
                # TODO maybe change this to '%' everywhere instead?
                'auto_id_prefix': 'id'
            }
        }
        implicit_settings.update(settings)
        settings = read_settings(path=None, override=implicit_settings)
        pelican = Pelican(settings=settings)
        pelican.run()

    def actual_expected_contents(self, actual, expected = None):
        if not expected: expected = actual

        with open(os.path.join(self.path, expected)) as f:
            expected_contents = f.read().strip()
        with open(os.path.join(self.path, 'output', actual)) as f:
            actual_contents = f.read().strip()
        return actual_contents, expected_contents



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

try:
	from poxy.utils import *
	import poxy.run as run
except:
	from utils import *
	import run

import re
import argparse
import datetime
from schema import SchemaError



def _invoker(func, **kwargs):
	try:
		func(**kwargs)
	except WarningTreatedAsError as err:
		print(rf'Error: {err} (warning treated as error)', file=sys.stderr)
		sys.exit(1)
	except SchemaError as err:
		print(err, file=sys.stderr)
		sys.exit(1)
	except Error as err:
		print(rf'Error: {err}', file=sys.stderr)
		sys.exit(1)
	except Exception as err:
		print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
		sys.exit(-1)
	sys.exit(0)



def _run(invoker=True):
	"""
	The entry point when the library is invoked as `poxy`.
	"""
	if invoker:
		_invoker(_run, invoker=False)
		return

	args = argparse.ArgumentParser(
		description=r'Generate fancy C++ documentation.',
		formatter_class=argparse.RawTextHelpFormatter
	)
	args.add_argument(
		r'config',
		type=Path,
		nargs='?',
		default=Path('.'),
		help=r'path to poxy.toml or a directory containing it (default: %(default)s)'
	)
	args.add_argument(
		r'-v', r'--verbose',
		action=r'store_true',
		help=r"enable very noisy diagnostic output"
	)
	args.add_argument(
		r'--doxygen',
		type=Path,
		default=None,
		metavar=r'"path"',
		help=r"specify the Doxygen executable to use (default: find on system path)"
	)
	args.add_argument(
		r'--dry',
		action=r'store_true',
		help=r"do a 'dry run' only, stopping after emitting the effective Doxyfile"
	)
	args.add_argument(
		r'--mcss',
		type=Path,
		default=None,
		metavar=r'"path"',
		help=r"specify the version of m.css to use (default: uses the bundled one)"
	)
	args.add_argument(
		r'--threads',
		type=int,
		default=0,
		metavar=r'N',
		help=r"set the number of threads to use (default: automatic)"
	)
	args.add_argument(
		r'--version',
		action=r'store_true',
		help=r"print the version and exit"
	)
	args.add_argument(
		r'--werror',
		action=r'store_true',
		help=r"always treat warnings as errors regardless of config file settings"
	)
	args.add_argument(
		r'--xmlonly',
		action=r'store_true',
		help=r"stop after generating and preprocessing the Doxygen xml"
	)
	args.add_argument(
		r'--ppinclude',
		type=str,
		default=None,
		metavar=r'"regex"',
		help=r"pattern matching HTML file names to post-process (default: all)"
	)
	args.add_argument(
		r'--ppexclude',
		type=str,
		default=None,
		metavar=r'"regex"',
		help=r"pattern matching HTML file names to exclude from post-processing (default: none)"
	)
	args.add_argument(r'--nocleanup', action=r'store_true', help=argparse.SUPPRESS)
	args = args.parse_args()

	if args.version:
		print(r'.'.join(lib_version()))
		return

	with ScopeTimer(r'All tasks', print_start=False, print_end=not args.dry) as timer:
		run(
			config_path = args.config,
			output_dir = Path.cwd(),
			threads = args.threads,
			cleanup = not args.nocleanup,
			verbose = args.verbose,
			mcss_dir = args.mcss,
			doxygen_path = args.doxygen,
			logger=True, # stderr + stdout
			dry_run=args.dry,
			xml_only=args.xmlonly,
			html_include=args.ppinclude,
			html_exclude=args.ppexclude,
			treat_warnings_as_errors=True if args.werror else None
		)



def _make_blog_post(invoker=True):
	"""
	The entry point when the library is invoked as `poxyblog`.
	"""
	if invoker:
		_invoker(_make_blog_post, invoker=False)
		return

	args = argparse.ArgumentParser(
		description=r'Initializes a new blog post for Poxy sites.',
		formatter_class=argparse.RawTextHelpFormatter
	)
	args.add_argument(
		r'title',
		type=str,
		help=r'the title of the new blog post'
	)
	args.add_argument(
		r'-v', r'--verbose',
		action=r'store_true',
		help=r"enable very noisy diagnostic output"
	)
	args.add_argument(
		r'--version',
		action=r'store_true',
		help=r"print the version and exit",
		dest=r'print_version'
	)
	args = args.parse_args()

	if args.print_version:
		print(r'.'.join(lib_version()))
		return

	date = datetime.datetime.now().date()

	title = args.title.strip()
	if not title:
		raise Error(r'title cannot be blank.')
	if re.search(r''''[\n\v\f\r]''', title) is not None:
		raise Error(r'title cannot contain newline characters.')
	file = re.sub(r'''[!@#$%^&*;:'"<>?/\\\s|+]+''', '_', title)
	file = rf'{date}_{file.lower()}.md'

	blog_dir = Path(r'blog')
	if blog_dir.exists() and not blog_dir.is_dir():
		raise Error(rf'{blog_dir.resolve()} already exists and is not a directory')
	blog_dir.mkdir(exist_ok=True)

	file = Path(blog_dir, file)
	if file.exists():
		if not file.is_file():
			raise Error(rf'{file.resolve()} already exist and is not a file')
		raise Error(rf'{file.resolve()} already exists')

	with open(file, r'w', encoding=r'utf-8', newline='\n') as f:
		write = lambda s='', end='\n': print(s, end=end, file=f)
		write(rf'# {title}')
		write()
		write()
		write()
		write()
		write()
		write(r'<!--[poxy_metadata[')
		write(rf'tags = []')
		write(r']]-->')
	print(rf'Blog post file initialized: {file.resolve()}')



if __name__ == '__main__':
	_run()



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

try:
	from poxy.utils import *
	import poxy.soup as soup
except:
	from utils import *
	import soup

import html
import json



#=======================================================================================================================
# base classes
#=======================================================================================================================

class HTMLFixer(object):
	pass

class PlainTextFixer(object):
	pass


#=======================================================================================================================
# custom tags
#=======================================================================================================================

class CustomTags(HTMLFixer):
	'''
	Modifies HTML using custom square-bracket [tags].
	'''
	__double_tags = re.compile(
		r'\[\s*('
			+ r'p|center|span|div|aside|code|pre|h1|h2|h3|h4|h5|h6|em|strong|b|i|u|li|ul|ol'
			+ r')(.*?)\s*\](.*?)\[\s*/\1\s*\]',
		re.I | re.S
	)
	__single_tags = re.compile(
		r'\[\s*(/?(?:'
			+ r'p|img|span|div|aside|code|pre|emoji'
			+ r'|(?:parent_)?set_(?:parent_)?(?:name|class)'
			+ r'|(?:parent_)?(?:add|remove)_(?:parent_)?class'
			+ r'|br|li|ul|ol|(?:html)?entity)'
			+ r')(\s+[^\]]+?)?\s*\]',
		re.I | re.S
	)
	__allowed_parents = ('dd', 'p', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'li', 'aside', 'td')

	@classmethod
	def __double_tags_substitute(cls, m, out, context):
		return f'<{m[1]}{html.unescape(m[2])}>{m[3]}</{m[1]}>'

	@classmethod
	def __single_tags_substitute(cls, m, out, context):
		tag_name = m[1].lower()
		tag_content = m[2].strip() if m[2] else ''
		if tag_name == 'htmlentity' or tag_name == 'entity':
			if not tag_content:
				return ''
			try:
				cp = int(tag_content, 16)
				if cp <= 0x10FFFF:
					return f'&#x{cp:X};'
			except:
				pass
			return f'&{tag_content};'
		elif tag_name == 'emoji':
			tag_content = tag_content.lower()
			if not tag_content:
				return ''
			for base in (16, 10):
				try:
					cp = int(tag_content, base)
					if cp in context.emoji_codepoints:
						return f'&#x{cp:X};&#xFE0F;'
				except:
					pass
			if tag_content in context.emoji:
				cp = context.emoji[tag_content][0]
				return f'&#x{cp:X};&#xFE0F;'
			return ''
		elif tag_name in (
				r'add_class', r'remove_class', r'set_class',
				r'parent_add_class', r'parent_remove_class', r'parent_set_class',
				r'add_parent_class', r'remove_parent_class', r'set_parent_class'):
			classes = []
			if tag_content:
				for s in tag_content.split():
					if s:
						classes.append(s)
			if classes:
				out.append((tag_name, classes))
			return ''
		elif tag_name in (r'set_name', r'parent_set_name', r'set_parent_name'):
			if tag_content:
				out.append((tag_name, tag_content))
			return ''
		else:
			return f'<{m[1]}{(" " + tag_content) if tag_content else ""}>'

	def __call__(self, doc, context):
		if doc.article_content is None:
			return False
		changed = False
		changed_this_pass = True
		while changed_this_pass:
			changed_this_pass = False
			for name in self.__allowed_parents:
				tags = doc.article_content.find_all(name)
				for tag in tags:
					if tag.decomposed or len(tag.contents) == 0 or soup.find_parent(tag, 'a', doc.article_content) is not None:
						continue
					replacer = RegexReplacer(self.__double_tags, lambda m, out: self.__double_tags_substitute(m, out, context), str(tag))
					if replacer:
						changed_this_pass = True
						soup.replace_tag(tag, str(replacer))
						continue
					replacer = RegexReplacer(self.__single_tags, lambda m, out: self.__single_tags_substitute(m, out, context), str(tag))
					if replacer:
						changed_this_pass = True
						parent = tag.parent
						new_tags = soup.replace_tag(tag, str(replacer))
						for i in range(len(replacer)):
							if replacer[i][0].find(r'parent_') != -1:
								if parent is None:
									continue
								if replacer[i][0] in (r'parent_add_class', r'add_parent_class'):
									soup.add_class(parent, replacer[i][1])
								elif replacer[i][0] in (r'parent_remove_class', r'remove_parent_class'):
									soup.remove_class(parent, replacer[i][1])
								elif replacer[i][0] in (r'parent_set_class', r'set_parent_class'):
									soup.set_class(parent, replacer[i][1])
								elif replacer[i][0] in (r'parent_set_name', r'set_parent_name'):
									parent.name = replacer[i][1]
							elif len(new_tags) == 1 and not isinstance(new_tags[0], soup.NavigableString):
								if replacer[i][0] == r'add_class':
									soup.add_class(new_tags[0], replacer[i][1])
								elif replacer[i][0] == r'remove_class':
									soup.remove_class(new_tags[0], replacer[i][1])
								elif replacer[i][0] == r'set_class':
									soup.set_class(new_tags[0], replacer[i][1])
								elif replacer[i][0] == r'set_name':
									new_tags[0].name = replacer[i][1]

						continue
			if changed_this_pass:
				doc.smooth()
				changed = True
		return changed



#=======================================================================================================================
# C++
#=======================================================================================================================

class _ModifiersBase(HTMLFixer):
	'''
	Base type for modifier parsing fixers.
	'''
	_modifierRegex = r"defaulted|noexcept|constexpr|(?:pure )?virtual|protected|__(?:(?:vector|std|fast)call|cdecl)"
	_modifierClasses = {
		"defaulted" : "m-info",
		"noexcept" : "m-success",
		"constexpr" : "m-primary",
		"pure virtual" : "m-warning",
		"virtual" : "m-warning",
		"protected" : "m-warning",
		"__vectorcall" : "m-special",
		"__stdcall" : "m-special",
		"__fastcall" : "m-special",
		"__cdecl" : "m-special"
	}



class Modifiers1(_ModifiersBase):
	'''
	Fixes improperly-parsed modifiers on function signatures in the various 'detail view' sections.
	'''
	__expression = re.compile(rf'(\s+)({_ModifiersBase._modifierRegex})(\s+)')
	__sections = ('pub-static-methods', 'pub-methods', 'friends', 'func-members')

	@classmethod
	def __substitute(cls, m, out):
		return f'{m[1]}<span class="poxy-injected m-label m-flat {cls._modifierClasses[m[2]]}">{m[2]}</span>{m[3]}'

	def __call__(self, doc, context):
		if doc.article_content is None:
			return False
		changed = False
		for sect in self.__sections:
			tags = doc.find_all_from_sections('dt', select='span.m-doc-wrap', section=sect)
			for tag in tags:
				replacer = RegexReplacer(self.__expression, self.__substitute, str(tag))
				if (replacer):
					changed = True
					soup.replace_tag(tag, str(replacer))
		return changed



class Modifiers2(_ModifiersBase):
	'''
	Fixes improperly-parsed modifiers on function signatures in the 'Function documentation' section.
	'''
	__expression = re.compile(rf'\s+({_ModifiersBase._modifierRegex})\s+')

	@classmethod
	def __substitute(cls, m, matches):
		matches.append(m[1])
		return ' '

	def __call__(self, doc, context):
		if doc.article_content is None:
			return False
		changed = False
		sections = doc.find_all_from_sections(section=False) # all sections without an id
		section = None
		for s in sections:
			if (str(s.h2.string) == 'Function documentation'):
				section = s
				break
		if (section is not None):
			funcs = section(id=True)
			funcs = [f.find('h3') for f in funcs]
			for f in funcs:
				bumper = f.select_one('span.m-doc-wrap-bumper')
				end = f.select_one('span.m-doc-wrap').contents
				end = end[len(end)-1]
				matches = []
				bumperContent = self.__expression.sub(lambda m: self.__substitute(m, matches), str(bumper))
				if (matches):
					changed = True
					soup.replace_tag(bumper, bumperContent)
					lastInserted = end.find('span')
					for match in matches:
						lastInserted = doc.new_tag('span',
							parent=end,
							string=match,
							class_=f'poxy-injected m-label {self._modifierClasses[match]}',
							before=lastInserted
						)
						lastInserted.insert_after(' ')
		return changed



class TemplateTemplate(HTMLFixer):
	'''
	Spreads consecutive template <> declarations out over multiple lines.
	'''
	__expression = re.compile(r'(template&lt;.+?&gt;)\s+(template&lt;)', re.S)

	@classmethod
	def __substitute(cls, m):
		return f'{m[1]}<br>\n{m[2]}'

	def __call__(self, doc, context):
		changed = False
		for template in doc.body('div', class_='m-doc-template'):
			replacer = RegexReplacer(self.__expression, lambda m, out: self.__substitute(m), str(template))
			if replacer:
				soup.replace_tag(template, str(replacer))
				changed = True
		return changed



class StripIncludes(HTMLFixer):
	'''
	Strips #include <paths/to/headers.h> based on context.sources.strip_includes.
	'''
	def __call__(self, doc, context):
		if doc.article is None or not context.sources.strip_includes:
			return False
		changed = False
		for include_div in doc.article.find_all(r'div', class_=r'm-doc-include'):
			anchor = include_div.find('a', href=True, class_=r'cpf')
			if anchor is None:
				continue
			text = anchor.get_text()
			if not (text.startswith('<') and text.endswith('>')):
				continue
			text = text[1:-1].strip()
			for strip in context.sources.strip_includes:
				if len(text) < len(strip) or not text.startswith(strip):
					continue
				if len(text) == len(strip):
					soup.destroy_node(include_div)
				else:
					anchor.contents.clear()
					anchor.contents.append(soup.NavigableString(rf'<{text[len(strip):]}>'))
				changed = True
				break
		return changed



class ImplementationDetails(PlainTextFixer):
	'''
	Replaces implementation details with appropriate shorthands.
	'''
	__shorthands = (
		(r'POXY_IMPLEMENTATION_DETAIL_IMPL', r'<code class="m-note m-dim poxy-impl">/* ... */</code>'),
	)
	def __call__(self, doc, context):
		changed = False
		for shorthand, replacement in self.__shorthands:
			idx = doc[0].find(shorthand)
			while idx >= 0:
				doc[0] = doc[0][:idx] + replacement + doc[0][idx + len(shorthand):]
				changed = True
				idx = doc[0].find(shorthand)
		return changed



#=======================================================================================================================
# index.html
#=======================================================================================================================

class IndexPage(HTMLFixer):
	'''
	Applies some basic fixes to index.html
	'''
	def __call__(self, doc, context):
		if doc.article_content is None or doc.path.name.lower() != 'index.html':
			return False
		parent = doc.article_content
		banner = parent.find('img')
		if banner:
			banner = banner.extract()
			parent.find('h1').replace_with(banner)
			if context.badges:
				parent = doc.new_tag('div', class_='gh-badges', after=banner)
				for (alt, src, href) in context.badges:
					if alt is None and src is None and href is None:
						doc.new_tag('br', parent=parent)
					else:
						anchor = doc.new_tag('a', parent=parent, href=href, target='_blank')
						doc.new_tag('img', parent=anchor, src=src, alt=alt)
				soup.add_class(banner, 'main_page_banner')
			return True
		return False



#=======================================================================================================================
# <code> blocks
#=======================================================================================================================

class CodeBlocks(HTMLFixer):
	'''
	Fixes various issues and improves syntax highlighting in <code> blocks.
	'''
	__keywords = (
		r'alignas',
		r'alignof',
		r'bool',
		r'char',
		r'char16_t',
		r'char32_t',
		r'char8_t',
		r'class',
		r'const',
		r'consteval',
		r'constexpr',
		r'constinit',
		r'do',
		r'double',
		r'else',
		r'explicit',
		r'false',
		r'float',
		r'if',
		r'inline',
		r'int',
		r'long',
		r'mutable',
		r'noexcept',
		r'short',
		r'signed',
		r'sizeof',
		r'struct',
		r'template',
		r'true',
		r'typename',
		r'unsigned',
		r'void',
		r'wchar_t',
		r'while',
	)

	__ns_token_expr = re.compile(r'(?:::|[a-zA-Z_][a-zA-Z_0-9]*|::[a-zA-Z_][a-zA-Z_0-9]*|[a-zA-Z_][a-zA-Z_0-9]*::)')
	__ns_full_expr = re.compile(r'(?:::)?[a-zA-Z_][a-zA-Z_0-9]*(::[a-zA-Z_][a-zA-Z_0-9]*)*(?:::)?')

	@classmethod
	def __colourize_compound_def(cls, tags, context):
		assert tags
		assert tags[0].string != '::'
		assert len(tags) == 1 or tags[-1].string != '::'
		full_str = ''.join([tag.get_text() for tag in tags])

		if context.code_blocks.enums.fullmatch(full_str):
			soup.set_class(tags[-1], 'ne')
			del tags[-1]
			while tags and tags[-1].string == '::':
				del tags[-1]
			if tags:
				cls.__colourize_compound_def(tags, context)
			return True

		if context.code_blocks.types.fullmatch(full_str):
			soup.set_class(tags[-1], 'ut')
			del tags[-1]
			while tags and tags[-1].string == '::':
				del tags[-1]
			if tags:
				cls.__colourize_compound_def(tags, context)
			return True

		while not context.code_blocks.namespaces.fullmatch(full_str):
			del tags[-1]
			while tags and tags[-1].string == '::':
				del tags[-1]
			if not tags:
				break
			full_str = ''.join([tag.get_text() for tag in tags])

		if tags:
			while len(tags) > 1:
				tags.pop().decompose()
			tags[0].string = full_str
			if soup.remove_class(tags[0], ('n', 'nl', 'kt')):
				soup.add_class(tags[0], 'ns')
			return True

		return False

	@classmethod
	def __adjacent_maybe_by_whitespace(cls, a, b):
		mid = a.next_sibling
		if mid is b:
			return True
		if not mid.next_sibling is b:
			return False
		if not isinstance(mid, soup.NavigableString):
			return False
		if len(mid.string.strip()) > 0:
			return False
		return True

	def __call__(self, doc, context):
		changed = False

		# fix up syntax highlighting
		code_blocks = doc.body(('pre','code'), class_='m-code')
		changed_this_pass = True
		while changed_this_pass:
			changed_this_pass = False
			for code_block in code_blocks:
				changed_this_block = False

				# c-style multi-line comments (doxygen butchers them)
				mlc_open = code_block.find('span', class_='o', string='/!*')
				while mlc_open is not None:
					mlc_close = mlc_open.find_next_sibling('span', class_='o', string='*!/')
					if mlc_close is None:
						break
					changed_this_block = True
					next_open = mlc_close.find_next_sibling('span', class_='o', string='/!*')

					tags = []
					current = mlc_open
					while current is not None:
						tags.append(current)
						if current is mlc_close:
							break
						current = current.next_sibling

					mlc_open.string = '/*'
					mlc_close.string = '*/'
					string = ''
					for tag in tags:
						string = string + tag.get_text()
					mlc_open.string = string
					soup.set_class(mlc_open, 'cm')
					while len(tags) > 1:
						soup.destroy_node(tags.pop())

					mlc_open = next_open

				# collect all names and glom them all together as compound names
				spans = code_block('span', class_=('n', 'nl', 'kt'), string=True)
				compound_names = []
				compound_name_evaluated_tags = set()
				for i in range(0, len(spans)):

					current = spans[i]
					if id(current) in compound_name_evaluated_tags:
						continue

					compound_name_evaluated_tags.add(id(current))
					tags = [ current ]
					while True:
						prev = current.previous_sibling
						if (prev is None
							or prev.string is None
							or isinstance(prev, soup.NavigableString)
							or 'class' not in prev.attrs
							or prev['class'][0] not in ('n', 'nl', 'kt', 'o')
							or not self.__ns_token_expr.fullmatch(prev.string)):
							break
						current = prev
						tags.insert(0, current)
						compound_name_evaluated_tags.add(id(current))

					current = spans[i]
					while True:
						nxt = current.next_sibling
						if (nxt is None
							or nxt.string is None
							or isinstance(nxt, soup.NavigableString)
							or 'class' not in nxt.attrs
							or nxt['class'][0] not in ('n', 'nl', 'kt', 'o')
							or not self.__ns_token_expr.fullmatch(nxt.string)):
							break
						current = nxt
						tags.append(current)
						compound_name_evaluated_tags.add(id(current))

					full_str = ''.join([tag.get_text() for tag in tags])
					if self.__ns_full_expr.fullmatch(full_str):
						while tags and tags[0].string == '::':
							del tags[0]
						while tags and tags[-1].string == '::':
							del tags[-1]
						if tags:
							compound_names.append(tags)

				# types and namespaces
				for tags in compound_names:
					if self.__colourize_compound_def(tags, context):
						changed_this_block = True

				# string and numeric literals
				spans = code_block('span', class_='n', string=True)
				for span in spans:
					prev = span.previous_sibling
					if (prev is None
						or isinstance(prev, soup.NavigableString)
						or 'class' not in prev.attrs):
						continue
					if ('s' in prev['class'] and context.code_blocks.string_literals.fullmatch(span.get_text())):
						soup.set_class(span, 'sa')
						changed_this_block = True
					elif (prev['class'][0] in ('mf', 'mi', 'mb', 'mh') and context.code_blocks.numeric_literals.fullmatch(span.get_text())):
						soup.set_class(span, prev['class'][0])
						changed_this_block = True

				# preprocessor macros
				spans = code_block('span', class_=('n', 'nl', 'kt', 'nc', 'nf'), string=True)
				for span in spans:
					if context.code_blocks.macros.fullmatch(span.get_text()):
						soup.set_class(span, r'm')
						changed_this_block = True

				# misidentifed keywords
				spans = code_block('span', class_=('nf', 'nb', 'kt', 'ut', 'kr'), string=True)
				for span in spans:
					if (span.string in self.__keywords):
						soup.set_class(span, r'k')
						changed_this_block = True

				# 'using' statements
				if 1:
					spans = code_block(r'span', class_=r'k', string=r'using')
					for using in spans:
						next_identifier = using.find_next_sibling(r'span', class_=r'n', string=True)
						if next_identifier is None:
							continue
						next_assign = next_identifier.find_next_sibling(r'span', class_=r'o', string=r'=')
						if next_assign is None:
							continue
						if not (self.__adjacent_maybe_by_whitespace(using, next_identifier)
								and self.__adjacent_maybe_by_whitespace(next_identifier, next_assign)):
							continue
						soup.set_class(next_identifier, r'ut')
						changed_this_block = True

				if changed_this_block:
					code_block.smooth()
					changed_this_pass = True
			changed = changed or changed_this_pass

		# fix doxygen butchering code blocks as inline nonsense
		code_blocks = doc.body('code', class_=('m-code', 'm-console'))
		changed_this_pass = True
		while changed_this_pass:
			changed_this_pass = False
			for code_block in code_blocks:
				parent = code_block.parent
				if (parent is None
					or parent.name != r'p'
					or parent.parent is None
					or parent.parent.name not in (r'div', r'section')):
					continue
				changed_this_pass = True
				code_block.name = 'pre'
				parent.insert_before(code_block.extract())
				parent.smooth()
				if (not parent.contents
					or (len(parent.contents) == 1
						and parent.contents[0].string.strip() == '')):
					soup.destroy_node(parent)
			changed = changed or changed_this_pass

		return changed



#=======================================================================================================================
# <a> tags
#=======================================================================================================================

def _m_doc_anchor_tags(tag):
	return (tag.name == 'a'
		and tag.has_attr('class')
		and ('m-doc' in tag['class']) # or 'm-doc-self' in tag['class'])
		and (tag.string is not None or tag.strings is not None)
	)



class AutoDocLinks(HTMLFixer):
	'''
	Adds links to additional sources where appropriate.
	'''
	__allowedNames = ('dd', 'p', 'dt', 'h3', 'td', 'div', 'figcaption')

	@classmethod
	def __substitute(cls, m, uri):
		external = uri.startswith('http')
		return rf'''<a href="{uri}" class="m-doc poxy-injected{' poxy-external' if external else ''}"{' target="_blank"' if external else ''}>{m[0]}</a>'''

	def __call__(self, doc, context):
		if doc.article_content is None:
			return False

		changed = False

		# first check all existing doc links to make sure they aren't erroneously linked to the wrong thing
		if 1:
			existing_doc_links = doc.article_content.find_all(_m_doc_anchor_tags)
			for link in existing_doc_links:
				done = False
				s = link.get_text()
				for expr, uri in context.autolinks:
					# check that it's a match for the replacement expression
					if not expr.fullmatch(s):
						continue
					# check the existing href against the target first
					if link.has_attr('href'):
						href = str(link['href'])
						anchor = href.rfind('#')
						if anchor == 0:
							continue # don't override internal self-links
						if anchor != -1:
							href = href[:anchor]
						if href == uri or href == doc.path.name: # don't override internal self-links
							continue
					link['href'] = uri
					soup.set_class(link, ['m-doc', 'poxy-injected'])
					if uri.startswith('http'):
						soup.add_class(link, 'poxy-external')
					done = True
					changed = True
					break
				if done:
					continue

		# now search the document for any other potential links
		if 1:
			tags = soup.shallow_search(doc.article_content, self.__allowedNames, lambda t: soup.find_parent(t, 'a', doc.article_content) is None)
			strings = []
			for tag in tags:
				strings = strings + soup.string_descendants(tag, lambda t: soup.find_parent(t, 'a', tag) is None)
			strings = [s for s in strings if s.parent is not None]
			for expr, uri in context.autolinks:
				if uri == doc.path.name: # don't create unnecessary self-links
					continue
				i = 0
				while i < len(strings):
					string = strings[i]
					parent = string.parent
					replacer = RegexReplacer(expr, lambda m, out: self.__substitute(m, uri), html.escape(str(string), quote=False))
					if replacer:
						repl_str = str(replacer)
						begins_with_ws = len(repl_str) > 0 and repl_str[:1].isspace()
						new_tags = soup.replace_tag(string, repl_str)
						if (begins_with_ws and new_tags[0].string is not None and not new_tags[0].string[:1].isspace()):
							new_tags[0].insert_before(' ')
						changed = True
						del strings[i]
						for tag in new_tags:
							strings = strings + soup.string_descendants(tag, lambda t: soup.find_parent(t, 'a', parent) is None)
						continue
					i = i + 1
		return changed



class Links(HTMLFixer):
	'''
	Fixes various minor issues with anchor tags.
	'''
	__external_href = re.compile(r'^(?:https?|s?ftp|mailto)[:].+$', re.I)
	__internal_doc_id = re.compile(r'^[a-fA-F0-9]+$')
	__godbolt = re.compile(r'^\s*https[:]//godbolt.org/z/.+?$', re.I)
	__local_href = re.compile(r'^([-/_a-zA-Z0-9]+\.[a-zA-Z]+)(?:#(.*))?$')

	def __call__(self, doc, context):
		changed = False
		for anchor in doc.body('a', href=True):
			href = anchor['href']

			# make sure links to external sources are correctly marked as such
			if self.__external_href.fullmatch(href) is not None:
				if 'target' not in anchor.attrs or anchor['target'] != '_blank':
					anchor['target'] = '_blank'
					changed = True
				changed = soup.add_class(anchor, 'poxy-external') or changed

				# do magic with godbolt.org links
				if self.__godbolt.fullmatch(href):
					changed = soup.add_class(anchor, 'godbolt') or changed
					if (anchor.parent.name == 'p'
							and len(anchor.parent.contents) == 1
							and anchor.parent.next_sibling is not None
							and anchor.parent.next_sibling.name == 'pre'):
						soup.add_class(anchor.parent, ('m-note', 'm-success', 'godbolt'))
						code_block = anchor.parent.next_sibling
						code_block.insert(0, anchor.parent.extract())
						changed = True
				continue

			is_mdoc = r'class' in anchor.attrs and (r'm-doc' in anchor['class'] or r'm-doc-self' in anchor['class'])

			# make sure links to local files point to actual existing files
			match = self.__local_href.fullmatch(href)
			if match and not coerce_path(doc.path.parent, match[1]).exists():
				changed = True
				if is_mdoc:
					href = r'#'
					anchor[r'href'] = r'#' # will by fixed by the next step
				else:
					for attr in (r'download', r'href', r'hreflang', r'media', r'ping', r'referrerpolicy', r'rel', r'target', r'type'):
						if attr in anchor.attrs:
							del anchor[attr]
					anchor.name = r'span'
					continue


			# make sure internal documentation #id links actually have somewhere to go
			if (is_mdoc
					and href.startswith(r'#')
					and (len(href) == 1 or doc.body.find(id=href[1:]) is None)):
				changed = True
				soup.remove_class(anchor, 'm-doc')
				soup.add_class(anchor, 'm-doc-self')
				anchor['href'] = '#'
				parent_with_id = anchor.find_parent(id=self.__internal_doc_id)
				if parent_with_id is None:
					parent_with_id = anchor.find_parent((r'dt', r'tr'), id=False)
					if parent_with_id is not None:
						parent_with_id['id'] = sha256(parent_with_id.get_text())
				if parent_with_id is None:
					parent_with_id = anchor.find_parent(id=True)
				if parent_with_id is not None:
					anchor['href'] = '#' + parent_with_id['id']
				continue

		return changed



#=======================================================================================================================
# empty tags
#=======================================================================================================================

class EmptyTags(HTMLFixer):
	'''
	Prunes the tree of various empty tags (happens as a side-effect of some other operations).
	'''
	def __call__(self, doc, context):
		changed = False
		for tag in doc.body((r'p', r'span')):
			if not tag.contents or (len(tag.contents) == 1 and isinstance(tag.contents[0], soup.NavigableString) and not tag.string):
				soup.destroy_node(tag)
				changed = True
		return changed



#=======================================================================================================================
# <head> tags
#=======================================================================================================================

class HeadTags(HTMLFixer):
	'''
	Injects poxy-specific tags into the document's <head> block.
	'''

	@classmethod
	def __append(cls, doc, name, attrs):
		doc.head.append('  ')
		tag = doc.new_tag(name, parent=doc.head, attrs=attrs)
		doc.head.append('\n')
		return tag

	def __call__(self, doc, context):

		# <meta> tags
		if 1:
			meta = []

			# name
			if context.name:
				if r'twitter:title' not in context.meta_tags:
					meta.append({ r'name' : r'twitter:title', r'content' : context.name})
				meta.append({ r'property' : r'og:title', r'content' : context.name})
				meta.append({ r'itemprop' : r'name', r'content' : context.name})

			# author
			if context.author:
				if r'author' not in context.meta_tags:
					meta.append({ r'name' : r'author', r'content' : context.author})
				meta.append({ r'property' : r'article:author', r'content' : context.author})

			# description
			if context.description:
				if r'description' not in context.meta_tags:
					meta.append({ r'name' : r'description', r'content' : context.description})
				if r'twitter:description' not in context.meta_tags:
					meta.append({ r'name' : r'twitter:description', r'content' : context.description})
				meta.append({ r'property' : r'og:description', r'content' : context.description})
				meta.append({ r'itemprop' : r'description', r'content' : context.description})

			# robots
			if not context.robots:
				if r'robots' not in context.meta_tags:
					meta.append({ r'name' : r'robots', r'content' : r'noindex, nofollow'})
				if r'googlebot' not in context.meta_tags:
					meta.append({ r'name' : r'googlebot', r'content' : r'noindex, nofollow'})

			# misc
			if r'format-detection' not in context.meta_tags:
				meta.append({ r'name' : r'format-detection', r'content' : r'telephone=no'})
			if r'generator' not in context.meta_tags:
				meta.append({ r'name' : r'generator', r'content' : rf'Poxy v{context.version_string}'})
			if r'referrer' not in context.meta_tags:
				meta.append({ r'name' : r'referrer', r'content' : r'no-referrer-when-downgrade'})

			# additional user-specified meta tags
			for name, content in context.meta_tags.items():
				meta.append({ r'name' : name, r'content' : content})

			for tag in meta:
				self.__append(doc, r'meta', tag)

		# google structured data
		if doc.path.name in context.compound_pages and context.compound_pages[doc.path.name][r'kind'] == r'page':
			tag = self.__append(doc, r'script', { r'type' : r'application/ld+json' })
			data = {}
			data[r'@context'] = r'https://schema.org'
			data[r'@type'] = r'Article'
			data[r'dateModified'] = context.now.isoformat()
			data[r'headline'] = doc.head.title.string
			tag.string = json.dumps(data, indent=r'    ')

		# stylesheets
		for stylesheet in context.stylesheets:
			self.__append(doc, r'link', { r'href' :  stylesheet, r'rel' : r'stylesheet' })

		# scripts
		for script in context.scripts:
			self.__append(doc, r'script', { r'src' : script })


		return True



#=======================================================================================================================
# misc
#=======================================================================================================================

class MarkTOC(HTMLFixer):
	'''
	Marks any table-of-contents with a custom class.
	'''
	def __call__(self, doc, context):
		if doc.table_of_contents is None:
			return False
		soup.add_class(doc.table_of_contents, r'poxy-toc')
		return True



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

import sys
import re
import io
import logging
from pathlib import Path
from io import StringIO
from misk import *


#=======================================================================================================================
# FUNCTIONS
#=======================================================================================================================

def coerce_path(arg, *args):
	assert arg is not None
	if args is not None and len(args):
		return Path(str(arg), *[str(a) for a in args])
	else:
		if not isinstance(arg, Path):
			arg = Path(str(arg))
		return arg



def coerce_collection(val):
	assert val is not None
	if not is_collection(val):
		val = ( val, )
	return val



def regex_or(patterns, pattern_prefix = '', pattern_suffix = '', flags=0):
	patterns = [str(r) for r in patterns if r is not None and r]
	patterns.sort()
	pattern = ''
	if patterns:
		pattern = '(?:(?:' + ')|(?:'.join(patterns) + '))'
	patterns = re.compile(pattern_prefix + pattern + pattern_suffix, flags=flags)
	return patterns



def log(logger, msg, level=logging.INFO):
	if logger is None or msg is None:
		return
	if isinstance(logger, bool):
		if logger:
			print(msg, file=sys.stderr if level >= logging.WARNING else sys.stdout)
	elif isinstance(logger, logging.Logger):
		logger.log(level, msg)
	elif isinstance(logger, io.IOBase):
		print(msg, file=logger)
	else:
		logger(msg)



def enum_subdirs(root, filter=None, recursive=False):
	root = coerce_path(root)
	assert root.is_dir()
	subdirs = []
	for p in root.iterdir():
		if p.is_dir():
			if filter is not None and not filter(p):
				continue
			subdirs.append(p)
			if recursive:
				subdirs = subdirs + enum_subdirs(p, filter=filter, recursive=True)
	return subdirs



def combine_dicts(x, y):
	z = x.copy()
	z.update(y)
	return z



_is_uri_regex = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]*://.+$')
def is_uri(s):
	global _is_uri_regex
	return _is_uri_regex.fullmatch(str(s)) is not None



_lib_version = None
def lib_version():
	global _lib_version
	if _lib_version is None:
		data_dir = Path(Path(__file__).resolve().parent, r'data')
		with open(Path(data_dir, 'version.txt'), encoding='utf-8') as file:
			_lib_version = [v.strip() for v in file.read().strip().split('.')]
			_lib_version = [v for v in _lib_version if v]
			assert len(_lib_version) == 3
			_lib_version = tuple(_lib_version)
	return _lib_version



#=======================================================================================================================
# REGEX REPLACER
#=======================================================================================================================

class RegexReplacer(object):

	def __substitute(self, m):
		self.__result = True
		return self.__handler(m, self.__out_data)

	def __init__(self, regex, handler, value):
		self.__handler = handler
		self.__result = False
		self.__out_data = []
		self.__value = regex.sub(lambda m: self.__substitute(m), value)

	def __str__(self):
		return self.__value

	def __bool__(self):
		return self.__result

	def __len__(self):
		return len(self.__out_data)

	def __getitem__(self, index):
		return self.__out_data[index]



#=======================================================================================================================
# CppTree
#=======================================================================================================================

class CppTree(object):

	NAMESPACES = 1
	TYPES = 2
	ENUM_VALUES = 4

	class Node(object):

		def __init__(self, val, parent, type_ = 0):
			assert val.find(r'::') == -1
			assert type_ in (0, CppTree.NAMESPACES, CppTree.TYPES, CppTree.ENUM_VALUES)
			self.value = val
			self.parent = parent
			self.type = type_
			self.mask = type_
			self.children = {}

		def add(self, val, type_ = 0):
			assert val.find(r'::') == -1
			assert type_ in (0, CppTree.NAMESPACES, CppTree.TYPES, CppTree.ENUM_VALUES)
			child = None
			if val not in self.children:
				child = CppTree.Node(val, self, type_)
				self.children[val] = child
			else:
				child = self.children[val]
				if type_:
					assert child.type in (0, type_)
					child.type = type_
					child.mask = child.mask | type_
			self.mask = self.mask | child.mask
			return child

		def regex_matcher(self, type_):
			assert type_ in (CppTree.NAMESPACES, CppTree.TYPES, CppTree.ENUM_VALUES)
			if not (type_ & self.mask):
				return None
			if not self.children:
				return self.value
			matchers = [v.regex_matcher(type_) for k, v in self.children.items()]
			matchers = [v for v in matchers if v is not None]
			if not matchers:
				if self.type == type_:
					return self.value
				return None

			grouped = len(matchers) > 1
			matchers = r'|'.join(matchers)
			if not self.value and not self.parent: # root
				return matchers
			matchers = (r'(?:' if grouped else '') + matchers + (r')' if grouped else '')

			if self.type == type_:
				return rf'{self.value}(?:::{matchers})?'
			else:
				return rf'{self.value}::{matchers}'



	def __init__(self):
		self.root = CppTree.Node('', None)

	def add(self, val, type_):
		assert type_ in (CppTree.NAMESPACES, CppTree.TYPES, CppTree.ENUM_VALUES)
		val = [v for v in val.split(r'::') if len(v)]
		parent = self.root
		while len(val):
			v = val.pop(0)
			parent = parent.add(v, type_ if not len(val) else 0)

	def add_type(self, val):
		self.add(val, CppTree.TYPES)

	def add_namespace(self, val):
		self.add(val, CppTree.NAMESPACES)

	def add_enum_value(self, val):
		self.add(val, CppTree.ENUM_VALUES)

	def matcher(self, type_):
		return self.root.regex_matcher(type_)



#=======================================================================================================================
# Custom exceptions
#=======================================================================================================================

class Error(Exception):
	"""Base class for other exceptions."""

	def __init__(self, *message):
		self.__message = r' '.join([str(m) for m in message])
		super().__init__(*message)

	def __str__(self):
		return self.__message



class WarningTreatedAsError(Error):
	"""Raised when a warning is generated and the user has chosen to treat warnings as errors."""
	pass



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT

try:
	from poxy.utils import *
except:
	from utils import *

import subprocess



#=======================================================================================================================
# functions
#=======================================================================================================================

def mangle_name(name):
	'''
	A lightweight version of doxygen's escapeCharsInString()
	(see https://github.com/doxygen/doxygen/blob/master/src/util.cpp)
	'''
	assert name is not None
	name = name.replace('_', '__')
	name = name.replace(':', '_1')
	name = name.replace('/', '_2')
	name = name.replace('<', '_3')
	name = name.replace('>', '_4')
	name = name.replace('*', '_5')
	name = name.replace('&', '_6')
	name = name.replace('|', '_7')
	name = name.replace('.', '_8')
	name = name.replace('!', '_9')
	name = name.replace(',', '_00')
	name = name.replace(' ', '_01')
	name = name.replace('{', '_02')
	name = name.replace('}', '_03')
	name = name.replace('?', '_04')
	name = name.replace('^', '_05')
	name = name.replace('%', '_06')
	name = name.replace('(', '_07')
	name = name.replace(')', '_08')
	name = name.replace('+', '_09')
	name = name.replace('=', '_0a')
	name = name.replace('$', '_0b')
	name = name.replace('\\','_0c')
	name = name.replace('@', '_0d')
	name = name.replace(']', '_0e')
	name = name.replace('[', '_0f')
	name = name.replace('#', '_0g')
	name = re.sub(r'[A-Z]', lambda m: '_' + m[0].lower(), name)
	return name


def _format_for_doxyfile(val):
	if val is None:
		return ''
	elif isinstance(val, str):
		return '"' + val.replace('"','\\"') + '"'
	elif isinstance(val, Path):
		return _format_for_doxyfile(str(val))
	elif isinstance(val, bool):
		return r'YES' if val else r'NO'
	elif isinstance(val, (int, float)):
		return str(val)
	else:
		assert False



#=======================================================================================================================
# Doxyfile
#=======================================================================================================================

class Doxyfile(object):

	def __init__(self, doxyfile_path, cwd=None, logger=None, doxygen_path=None, flush_at_exit=True):
		self.__logger=logger
		self.__dirty=True
		self.__text = ''
		self.__autoflush=bool(flush_at_exit)

		# the path of the actual doxyfile
		self.path = coerce_path(doxyfile_path).resolve()

		# the working directory for doxygen invocations
		self.__cwd = Path.cwd() if cwd is None else coerce_path(cwd).resolve()
		assert_existing_directory(self.__cwd)

		# doxygen itself
		self.__doxygen = r'doxygen' if doxygen_path is None else coerce_path(doxygen_path)

		# read in doxyfile
		if self.path.exists():
			if not self.path.is_file():
				raise Error(f'{self.path} was not a file')
			self.__text = read_all_text_from_file(self.path, logger=self.__logger).strip()
			self.cleanup() # expands includes

		# ...or generate one
		else:
			result = subprocess.run(
				[str(self.__doxygen), r'-s', r'-g', r'-'],
				check=True,
				capture_output=True,
				cwd=self.__cwd,
				encoding='utf-8'
			)
			self.__text = result.stdout.strip()


		# simplify regex searches by ensuring there's always leading and trailing newlines
		self.__text = f'\n{self.__text}\n'

	def cleanup(self):
		if not self.__dirty:
			return
		if 1:
			log(self.__logger, rf'Invoking doxygen to clean doxyfile')
			result = subprocess.run(
				[str(self.__doxygen), r'-s', r'-u', r'-'],
				check=True,
				capture_output=True,
				cwd=self.__cwd,
				encoding=r'utf-8',
				input=self.__text
			)
			self.__text = result.stdout.strip()
		self.__dirty = False

	def flush(self):
		self.cleanup()
		log(self.__logger, rf'Writing {self.path}')
		with open(self.path, 'w', encoding='utf-8', newline='\n') as f:
			print(self.__text, file=f)

	def hash(self):
		return sha1(self.__text)

	def contains(self, text):
		assert text is not None
		return self.__text.find(text) != -1

	def get_value(self, key, fallback=None):
		pattern = re.compile(rf'\n\s*{key}\s*=(.*?)\n', flags=re.S)
		text = self.__text
		m = pattern.search(text)
		while m:
			# doxygen allows values to appear multiple times and only accepts the last one, hence the loop
			text = text[m.end():]
			n = pattern.search(text)
			if not n:
				break
			m = n
		if m:
			val = m[1].strip(' "')
			return val if val else fallback
		return fallback

	def get_boolean(self, key, fallback=False):
		val = self.get_value(key)
		if val is None:
			return fallback
		return val.upper() == 'YES'

	def append(self, *args, end='\n', sep=' '):
		self.__text = rf'{self.__text}{sep.join(args)}{end}'
		self.__dirty = True
		return self

	def add_value(self, key, value=None):
		if value is not None:
			if isinstance(value, (list, tuple, set)):
				if value:
					for v in value:
						if v is not None:
							self.append(rf'{key:<23}+= {_format_for_doxyfile(v)}')
			else:
				self.append(rf'{key:<23}+= {_format_for_doxyfile(value)}')
			self.__dirty = True
		return self

	def set_value(self, key, value=None):
		if isinstance(value, (list, tuple, set)):
			if not value:
				self.append(rf'{key:<23}=')
			else:
				first = True
				for v in value:
					if first:
						self.append(rf'{key:<23}=  {_format_for_doxyfile(v)}')
					else:
						self.add_value(key, v)
					first = False
		else:
			self.append(rf'{key:<23}=  {_format_for_doxyfile(value)}')
		self.__dirty = True
		return self

	def get_text(self):
		return self.__text

	def __enter__(self):
		return self

	def __exit__(self, type, value, traceback):
		if traceback is None and self.__autoflush:
			self.flush()
